__all__ = [
    "TypeForm",
    "type_form_repr",
]
TYPE_CHECKING = False
if TYPE_CHECKING:
    # TODO: Import from typing once PEP-747 is approved.
    # See https://peps.python.org/pep-0747/
    from typing_extensions import TypeForm

else:

    def __getattr__(name: str):
        match name:
            case "TypeForm":
                from typing_extensions import TypeForm

                return TypeForm
        raise AttributeError(f"Package '{__package__}' has no attribute '{name}'")


def type_form_repr(type_form: TypeForm) -> str:
    from tressed.predicates import get_args, is_union_type

    if name := getattr(type_form, "__name__", None):
        if type_params := getattr(type_form, "__type_params__", None):
            # C[T1=V1, .., Tn=?]
            params = []

            if args := get_args(type_form):
                num_args = len(args)
                for i, arg in enumerate(args):
                    params.append(
                        f"{type_form_repr(type_params[i])}={type_form_repr(arg)}"
                    )
            else:
                num_args = 0
            for type_param in type_params[num_args:]:
                params.append(f"{type_form_repr(type_param)}=?")
            return f"{name}[{', '.join(params)}]"

        if args := get_args(type_form):
            if len(args) > 1 and is_union_type(type_form):
                # T1 | .. | Tn
                return " | ".join(map(type_form_repr, args))
            # C[T1, .., Tn]
            return f"{name}[{', '.join(map(type_form_repr, args))}]"
        return name
    return repr(type_form)



TYPE_CHECKING = False
if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import Any, Protocol

    from tressed.type_path import TypePath

__all__ = ["Dumped", "DumperProtocol", "DumperFn"]


# Dumped type, representing types that can be serialized to json out of the box.
type Dumped = (
    int
    | float
    | str
    | bool
    | list[Dumped]
    | tuple[Dumped, ...]
    | dict[str, Dumped]
    | None
)

type DumperFn = Callable[[Any, TypePath, DumperProtocol], Dumped]

if TYPE_CHECKING:

    class DumperProtocol(Protocol):
        def _dump(self, value: Any, type_path: TypePath) -> Dumped: ...
        def _resolve_alias(
            self, value_type: type, type_path: TypePath, name: str
        ) -> str: ...

        @property
        def hide_defaults(self) -> bool: ...
else:
    # Placeholder type for runtime
    type DumperProtocol = type



import pytest

from tressed.dumper import Dumper
from tressed.exceptions import TressedTypeError


def test_dump_basic_types() -> None:
    dumper = Dumper()

    assert dumper.dump(1) == 1
    assert dumper.dump(1.23) == 1.23
    assert dumper.dump("foo") == "foo"
    assert dumper.dump(True) is True
    assert dumper.dump(False) is False
    assert dumper.dump(None) is None


def test_dump_simple_sequence() -> None:
    dumper = Dumper()

    assert dumper.dump([1, 2, 3]) == [1, 2, 3]
    assert dumper.dump((1, 2, 3)) == [1, 2, 3]

    dumped = dumper.dump({1, 2, 3})
    assert isinstance(dumped, list)
    assert sorted(dumped) == [1, 2, 3]  # type: ignore[type-var]

    dumped = dumper.dump(frozenset({1, 2, 3}))
    assert isinstance(dumped, list)
    assert sorted(dumped) == [1, 2, 3]  # type: ignore[type-var]


def test_dump_ipaddress() -> None:
    import ipaddress

    dumper = Dumper()

    assert dumper.dump(ipaddress.IPv4Address("127.0.0.1")) == "127.0.0.1"
    assert dumper.dump(ipaddress.IPv6Address("::1")) == "::1"

    assert dumper.dump(ipaddress.IPv4Network("127.0.0.0/24")) == "127.0.0.0/24"
    assert dumper.dump(ipaddress.IPv6Network("::0/64")) == "::/64"

    assert dumper.dump(ipaddress.IPv4Interface("127.0.0.1/24")) == "127.0.0.1/24"
    assert dumper.dump(ipaddress.IPv6Interface("::1/64")) == "::1/64"


def test_dump_enum() -> None:
    import enum

    class SampleEnum(enum.Enum):
        FOO = (1, 2)

    class SampleStrEnum(enum.StrEnum):
        BAR = enum.auto()

    class SampleIntEnum(enum.IntEnum):
        BAZ = enum.auto()

    dumper = Dumper()

    assert dumper.dump(SampleEnum.FOO) == [1, 2]
    assert dumper.dump(SampleStrEnum.BAR) == "bar"
    assert dumper.dump(SampleIntEnum.BAZ) == 1


def test_dump_datetime() -> None:
    from datetime import UTC, date, datetime, time, timedelta, timezone

    dumper = Dumper()
    assert dumper.dump(time(12, 45, 59)) == "12:45:59"
    assert dumper.dump(time(12, 45, 59, tzinfo=UTC)) == "12:45:59Z"
    assert (
        dumper.dump(time(12, 45, 59, tzinfo=timezone(timedelta(seconds=-1800))))
        == "12:45:59-00:30"
    )
    assert dumper.dump(date(2025, 12, 29)) == "2025-12-29"
    assert dumper.dump(datetime(2025, 12, 29, 12, 45, 59)) == "2025-12-29T12:45:59"
    assert (
        dumper.dump(datetime(2025, 12, 29, 12, 45, 59, tzinfo=UTC))
        == "2025-12-29T12:45:59Z"
    )
    assert (
        dumper.dump(
            datetime(2025, 12, 29, 12, 45, 59, tzinfo=timezone(timedelta(seconds=7200)))
        )
        == "2025-12-29T12:45:59+02:00"
    )


def test_dump_dataclass() -> None:
    import uuid
    from dataclasses import dataclass, field

    from tressed.alias import to_pascal

    @dataclass
    class SomeDataclass:
        foo: str
        some_secret: str = field(repr=False)
        bar: str = "bar"
        baz: list[int] = field(default_factory=lambda: [1, 2, 3])
        bar_bar: tuple[int, str] = field(metadata={"alias": "barBar"}, kw_only=True)
        some_id: uuid.UUID = field(default_factory=uuid.uuid4)
        some_lazy_field: int = field(init=False)

        def __post_init__(self) -> None:
            self.some_lazy_field = len(self.baz)

    value = SomeDataclass(
        foo="FOO",
        some_secret="hehe",
        bar="bar",
        bar_bar=(123, "Hej ho!"),
        some_id=uuid.UUID("c932f581-6430-4ae1-ad63-85489a0206b2"),
    )
    assert value.some_lazy_field == 3

    dumper = Dumper(alias_fn=to_pascal)
    assert dumper.dump(value) == {
        "Bar": "bar",
        "Baz": [1, 2, 3],
        "Foo": "FOO",
        "barBar": [123, "Hej ho!"],
        "SomeId": "c932f581-6430-4ae1-ad63-85489a0206b2",
    }

    dumper_hide_defaults = Dumper(alias_fn=to_pascal, hide_defaults=True)
    assert dumper_hide_defaults.dump(value) == {
        "Foo": "FOO",
        "barBar": [123, "Hej ho!"],
        "SomeId": "c932f581-6430-4ae1-ad63-85489a0206b2",
    }


def test_dump_namedtuple() -> None:
    from typing import NamedTuple

    from tressed.alias import to_camel

    class SomeNamedTuple(NamedTuple):
        some_field: int
        some_default_field: str = "foobar"

    value = SomeNamedTuple(some_field=123)

    dumper = Dumper(alias_fn=to_camel)
    assert dumper.dump(value) == {"someField": 123, "someDefaultField": "foobar"}

    dumper = Dumper(hide_defaults=True)
    assert dumper.dump(value) == {"some_field": 123}


def test_dump_unknown_type() -> None:
    class UnknownType:
        def __repr__(self) -> str:
            return f"{self.__class__.__name__}()"

    dumper = Dumper()

    with pytest.raises(TressedTypeError) as exc_info:
        dumper.dump(UnknownType())
    assert (
        str(exc_info.value)
        == "Unhandled type UnknownType at path . for value UnknownType()"
    )


def test_dump_custom_type() -> None:
    from typing import Any

    from tressed import TypeForm, TypePath
    from tressed.dumper import Dumped, DumperProtocol

    class CustomType:
        def __init__(self, value: int) -> None:
            self.value = value

    def _is_custom_type(type_form: TypeForm) -> bool:
        if type(type_form) is type:
            return issubclass(type_form, CustomType)
        else:
            return False

    def _dump_custom_type(
        value: Any, type_path: TypePath, dumper: DumperProtocol
    ) -> Dumped:
        return [f"Custom value is {value.value}", value.value]

    dumper = Dumper(extra_type_dumpers={CustomType: _dump_custom_type})
    assert dumper.dump(CustomType(123)) == ["Custom value is 123", 123]

    dumper = Dumper(extra_type_mappers={_is_custom_type: _dump_custom_type})
    assert dumper.dump(CustomType(123)) == ["Custom value is 123", 123]


def test_barebones_dumper() -> None:
    dumper = Dumper(
        default_type_dumpers={int: lambda v, _, __: "*" * v},
        default_type_mappers={lambda t: t is tuple: lambda v, _, __: len(v)},
    )
    assert dumper.dump(10) == "**********"
    assert dumper.dump(("oo", "bar", "baz")) == 3

    with pytest.raises(TressedTypeError) as exc_info:
        dumper.dump(10.6) == "Unhandled type float at path . for value 10.6"

    with pytest.raises(TressedTypeError) as exc_info:
        dumper.dump("foo")
    assert str(exc_info.value) == "Unhandled type str at path . for value 'foo'"


def test_dump_path() -> None:
    from pathlib import Path

    dumper = Dumper()
    assert dumper.dump(Path("/foo/bar")) == "/foo/bar"

    class CustomPath:
        def __init__(self, value: int) -> None:
            self.value = value

        def __fspath__(self) -> str:
            return "/".join(["foo"] * self.value)

    assert dumper.dump(CustomPath(3)) == "foo/foo/foo"



TYPE_CHECKING = False
if TYPE_CHECKING:
    from typing import Any

    from tressed.loader.types import (
        LoaderFn,
        LoaderProtocol,
        TypeLoaderSpecializer,
        TypePath,
    )
    from tressed.type_form import TypeForm

__all__ = ["SpecializingLoader"]


class SpecializingLoader:
    def __init__(self, loader: LoaderFn, specializer: TypeLoaderSpecializer) -> None:
        self._loader = loader
        self._specialized_loaders: dict[tuple[TypeForm, TypePath], Any] = {}
        self._specializer = specializer
        self._threshold: int = 3

    def __call__[T](
        self,
        value: Any,
        type_form: TypeForm[T],
        type_path: TypePath,
        loader: LoaderProtocol,
    ) -> T:
        key = (type_form, type_path)
        specialized_loader = self._specialized_loaders.get(key)
        if type(specialized_loader) is type(None) or type(specialized_loader) is int:
            if type(specialized_loader) is int and specialized_loader > 0:
                count = specialized_loader + 1
            else:
                count = 1
            self._specialized_loaders[key] = count
            specialized_loader = None

            # Specialize!
            if count > self._threshold:
                specializer_fn_code = self._specializer(*key)
                if specializer_fn_code is None:
                    # zero means we can't specialize this
                    self._specialized_loaders[key] = 0
                else:
                    globals_: dict[str, Any] = {}
                    locals_: dict[str, Any] = {}

                    exec(specializer_fn_code, globals_, locals_)

                    specialized_loader = locals_["__specialized_fn"]
                    self._specialized_loaders[key] = specialized_loader

        if specialized_loader is None:
            return self._loader(value, type_form, type_path, loader)
        else:
            return specialized_loader(value, loader)



__all__ = []

TYPE_CHECKING = False
if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import Any, Protocol

    from tressed.type_form import TypeForm
    from tressed.type_path import TypePath

    class LoaderProtocol(Protocol):
        def _load[T](
            self, value: Any, type_form: TypeForm[T], type_path: TypePath
        ) -> T: ...
        def _resolve_alias[T](
            self, type_form: TypeForm[T], type_path: TypePath, name: str
        ) -> str: ...

    type LoaderFn[T] = Callable[[Any, TypeForm[T], TypePath, LoaderProtocol], T]
    type TypeLoaderSpecializer[T] = Callable[[TypeForm[T], TypePath], str | None]

    __all__ += [
        "LoaderProtocol",
        "LoaderFn",
        "TypeLoaderSpecializer",
    ]



TYPE_CHECKING = False
if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import Any, NoReturn

    from tressed.type_form import TypeForm
    from tressed.type_path import TypePath

__all__ = [
    "TressedError",
    "TressedTypeError",
    "TressedTypeFormError",
    "TressedValueError",
]


class TressedError(Exception):
    pass


class TressedTypeError(TressedError, TypeError):
    def __init__(
        self,
        value: Any,
        type_path: TypePath,
        message: str = "",
    ) -> None:
        self.value = value
        self.type = type(value)
        self.type_path = type_path
        self.message = message
        super(TypeError, self).__init__(self.type)

    def __str__(self) -> str:
        from tressed.type_form import type_form_repr
        from tressed.type_path import type_path_repr

        return (
            f"Unhandled type {type_form_repr(self.type)} "
            f"at path {type_path_repr(self.type_path)} "
            f"for value {self.value!r}{f': {self.message}' if self.message else ''}"
        )


class TressedTypeFormError(TressedError, TypeError):
    def __init__(
        self, value: Any, type_form: TypeForm, type_path: TypePath, message: str = ""
    ) -> None:
        self.value = value
        self.type_form = type_form
        self.type_path = type_path
        self.message = message
        super(TypeError, self).__init__(type_form)

    def __str__(self) -> str:
        from tressed.type_form import type_form_repr
        from tressed.type_path import type_path_repr

        return (
            f"Unhandled type form {type_form_repr(self.type_form)} "
            f"at path {type_path_repr(self.type_path)} "
            f"for value {self.value!r}{f': {self.message}' if self.message else ''}"
        )


class TressedValueError(TressedError, ValueError):
    def __init__(
        self,
        value: Any,
        type_form: TypeForm,
        type_path: TypePath,
        message: str = "",
        exceptions: Sequence[TressedValueError] = (),
    ) -> None:
        self.value = value
        self.type_form = type_form
        self.type_path = type_path
        self.message = message
        self.exceptions = exceptions
        super(ValueError, self).__init__(value)

    def __str__(self) -> str:
        from tressed.type_form import type_form_repr
        from tressed.type_path import type_path_repr

        value = (
            f"Failed to load value of type {type_form_repr(type(self.value))} "
            f"at path {type_path_repr(self.type_path)} "
            f"into type form {type_form_repr(self.type_form)}"
        )
        if self.message:
            value += f": {self.message}"
        if self.exceptions:
            value += f" ({len(self.exceptions)} sub-exceptions)"
        return value

    def raise_exception_group(self) -> NoReturn:
        """
        Raise as exception group.
        """
        raise ExceptionGroup(
            str(self),
            self.exceptions,
        )



TYPE_CHECKING = False
if TYPE_CHECKING:
    from tressed.loader.loader import Loader
    from tressed.loader.types import LoaderFn, LoaderProtocol

__all__ = ["Loader", "LoaderProtocol", "LoaderFn"]


if not TYPE_CHECKING:

    def __getattr__(name: str) -> type[Loader]:
        match name:
            case "Loader":
                from tressed.loader.loader import Loader

                return Loader

            case "LoaderProtocol" | "LoaderFn":
                from tressed.loader import types

                return getattr(types, name)

            case _:
                raise AttributeError(
                    f"Package '{__package__}' has no attribute '{name}'"
                )



TYPE_CHECKING = False
if TYPE_CHECKING:
    from collections.abc import Callable, Sequence
    from typing import Any, Final, Literal

    from tressed.type_form import TypeForm


__all__ = ["Discriminator"]

if TYPE_CHECKING:
    type MatchFn = Callable[[Any, TypeForm], bool | int]
    type MatchStrategy = Literal["first-match", "best-match"]

    __all__ += ["MatchFn", "MatchStrategy"]


class Discriminator:
    __slots__ = ("match_fn", "strategy", "match")

    def __init__(
        self,
        match_fn: MatchFn,
        strategy: MatchStrategy = "first-match",
    ) -> None:
        self.match_fn: Final = match_fn
        self.strategy: Final = strategy
        match strategy:
            case "first-match":
                self.match = self.first_match
            case "best-match":
                self.match = self.best_match
            case _:
                assert False, f"Invalid match strategy {strategy}"

    def best_match(self, value: Any, *type_forms: TypeForm) -> TypeForm | None:
        best_matches: Sequence[TypeForm] = ()
        best_score = 0
        for type_form in type_forms:
            score = int(self.match_fn(value, type_form))
            if score < 1:
                continue
            if score > best_score:
                best_matches = (type_form,)
                best_score = score
            elif score == best_score:
                best_matches = (*best_matches, type_form)
        match best_matches:
            case [type_form]:
                return type_form
            case []:
                # no match
                return None
            case _:
                # ambiguous match
                return None

    def first_match(self, value: Any, *type_forms: TypeForm) -> TypeForm | None:
        for type_form in type_forms:
            if self.match_fn(value, type_form):
                return type_form
        return None

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.match_fn=!r})"

    def __eq__(self, other: Any) -> bool:
        return type(self) is type(other) and self.match == other.match_fn

    def __hash__(self) -> int:
        return id(type(self)) + hash(self.match_fn) + hash(self.strategy)



TYPE_CHECKING = False
if TYPE_CHECKING:
    from collections.abc import Callable

    from tressed.type_form import TypeForm
    from tressed.type_path import TypePath


__all__ = [
    "AliasResolver",
    "normalize_alias_fn",
    "identity_alias_fn",
    "make_maybe_dataclass_alias_fn",
    "compose_alias_fn",
    "to_camel",
    "to_pascal",
]

if TYPE_CHECKING:
    type Alias = str
    type SimpleAliasFn[AliasT = Alias] = Callable[[str], AliasT]
    type TypeFormAliasFn[AliasT = Alias] = Callable[[str, TypeForm], AliasT]
    type TypePathAliasFn[AliasT = Alias] = Callable[[str, TypeForm, TypePath], AliasT]
    type AliasFn[AliasT = Alias] = (
        SimpleAliasFn[AliasT] | TypeFormAliasFn[AliasT] | TypePathAliasFn[AliasT]
    )

    __all__ += [
        "Alias",
        "SimpleAliasFn",
        "TypeFormAliasFn",
        "TypePathAliasFn",
        "AliasFn",
    ]


def normalize_alias_fn[AliasT = Alias](
    alias_fn: SimpleAliasFn[AliasT] | TypeFormAliasFn[AliasT] | AliasFn[AliasT],
) -> TypePathAliasFn[AliasT]:
    """
    Normalize alias fn to a type path alias fn.
    """
    # Avoid importing heavy inspect module, do some manual introspection
    arg_count = alias_fn.__code__.co_argcount
    if hasattr(alias_fn, "__self__"):
        # Bound method, substract the self argument
        arg_count -= 1

    match arg_count:
        case 3:
            return alias_fn  # type: ignore[return-value]
        case 2:

            def _type_form_alias_fn_wrapper(
                name: str, type_form: TypeForm, type_path: TypePath
            ) -> AliasT:
                return alias_fn(name, type_form)  # type: ignore[call-arg]

            return _type_form_alias_fn_wrapper
        case 1:

            def _simple_alias_fn_wrapper(
                name: str, type_form: TypeForm, type_path: TypePath
            ) -> AliasT:
                return alias_fn(name)  # type: ignore[call-arg]

            return _simple_alias_fn_wrapper
        case _:
            assert False, "unreachable"


def identity_alias_fn(name: str) -> str:
    return name


def _snake_to_camel_pascal(
    name: str, first_fn: Callable[[str], str], rest_fn: Callable[[str], str]
) -> str:
    # Find trailing dashes
    for index, char in enumerate(name):
        if char != "_":
            break

    parts = []
    end = len(name) - 1
    start = 0
    while start < end:
        index = name.find("_", index)

        # No more underscore, return rest
        if index < 0:
            break

        num_underscores = 1
        while index + 1 <= end and name[index + 1] == "_":
            index += 1
            num_underscores += 1

        # One underscore and more to come, remove underscore
        if num_underscores == 1 and index + 1 < end:
            parts.append(name[start:index].title())
        # More than one underscore or nothing more to come, preserve underscores as is
        else:
            parts.append(name[start : index + 1].title())
        index += 1
        start = index
    parts.append(name[start:])

    match parts:
        case [first, *parts]:
            return f"{first_fn(first)}{''.join(map(rest_fn, parts))}"
        case []:
            return ""
        case _:
            assert False, "unreachable"


def to_pascal(name: str) -> str:
    return _snake_to_camel_pascal(name, str.title, str.title)


def to_camel(name: str) -> str:
    return _snake_to_camel_pascal(name, str.lower, str.title)


def make_maybe_dataclass_alias_fn(alias_field: str) -> AliasFn[Alias | None]:
    def _maybe_dataclass_alias_fn(
        name: str, type_form: TypeForm, type_path: TypePath
    ) -> Alias | None:
        if fields := getattr(type_form, "__dataclass_fields__", None):
            field = fields[name]
            return field.metadata.get(alias_field)
        return None

    return _maybe_dataclass_alias_fn


def compose_alias_fn(
    *alias_fns: AliasFn[Alias | None],
    default_alias_fn: AliasFn[Alias] = identity_alias_fn,
) -> TypePathAliasFn:
    """
    Create an alias function by composing several alias functions.
    The functions are tried in order until an alias is returned.
    If none of the function match, the alias is resolved using the default alias function.
    """
    normalized_alias_fns = [normalize_alias_fn(alias_fn) for alias_fn in alias_fns]
    normalized_default_alias_fn = normalize_alias_fn(default_alias_fn)

    def _composed_alias_fn(
        name: str, type_form: TypeForm, type_path: TypePath
    ) -> Alias:
        for alias_fn in normalized_alias_fns:
            if (alias := alias_fn(name, type_form, type_path)) is not None:
                return alias
        return normalized_default_alias_fn(name, type_form, type_path)

    return _composed_alias_fn


class AliasResolver:
    def __init__(
        self, alias_fn: AliasFn | None = None, cache_resolved_aliases: bool = True
    ) -> None:
        if alias_fn is None:
            alias_fn = identity_alias_fn
        self._alias_fn = normalize_alias_fn(alias_fn)
        self._cache: dict[tuple[str, TypeForm, TypePath], Alias] = {}
        self.resolve = (
            self._resolve_cached if cache_resolved_aliases else self._alias_fn
        )

    def _resolve_cached(
        self, name: str, type_form: TypeForm, type_path: TypePath
    ) -> Alias:
        cache_key = (name, type_form, type_path)
        if (alias := self._cache.get(cache_key)) is not None:
            return alias

        alias = self._alias_fn(name, type_form, type_path)

        self._cache[cache_key] = alias
        return alias



import os
import sys

TYPE_CHECKING = False
if TYPE_CHECKING:
    from typing import Any

    from tressed.dumper.types import Dumped, DumperProtocol
    from tressed.exceptions import TressedValueError
    from tressed.type_path import TypePath

__all__ = [
    "dump_identity",
    "dump_simple_sequence",
    "dump_simple_mapping",
    "dump_simple_scalar",
    "dump_complex",
    "dump_enum",
    "dump_datetime",
    "dump_dataclass",
    "dump_namedtuple",
    "dump_fspath",
]

if TYPE_CHECKING:
    from enum import Enum, auto

    class _MissingType(Enum):
        _MISSING = auto()

    _MISSING = _MissingType._MISSING

else:
    _MISSING = object()


def dump_identity(value: Any, type_path: TypePath, dumper: DumperProtocol) -> Dumped:
    return value


def dump_simple_scalar(
    value: Any, type_path: TypePath, dumper: DumperProtocol
) -> Dumped:
    return str(value)


def dump_simple_sequence(
    value: Any, type_path: TypePath, dumper: DumperProtocol
) -> Dumped:
    return [dumper._dump(item, (*type_path, i)) for i, item in enumerate(value)]


def dump_complex(value: Any, type_path: TypePath, dumper: DumperProtocol) -> Dumped:
    return [value.real, value.imag]


def _dump_key(key: Any, type_path: TypePath, dumper: DumperProtocol) -> str:
    key_path = (*type_path, key)
    dumped_key = dumper._dump(key, key_path)
    if type(dumped_key) is not str:
        raise TressedValueError(key, type(key), key_path, "dumped key must be a string")
    return dumped_key


def dump_simple_mapping(
    value: Any, type_path: TypePath, dumper: DumperProtocol
) -> Dumped:
    return {
        _dump_key(item_key, type_path, dumper): dumper._dump(
            item_value, (*type_path, item_key)
        )
        for item_key, item_value in value.items()
    }


def dump_enum(value: Any, type_path: TypePath, dumper: DumperProtocol) -> Dumped:
    return dumper._dump(value.value, type_path)


def dump_datetime(value: Any, type_path: TypePath, dumper: DumperProtocol) -> Dumped:
    dumped = value.isoformat()
    # Return datetime.{date,datetime,time}.isoformat() with the one difference that we use
    # a Z suffix instead of +00:00 to keep the output more compact and more like the usual datetimes
    # seen in json payloads.
    zero_suffix = "+00:00"
    if dumped.endswith(zero_suffix):
        return dumped[: -len(zero_suffix)] + "Z"
    return dumped


def dump_dataclass(value: Any, type_path: TypePath, dumper: DumperProtocol) -> Dumped:
    from dataclasses import MISSING, fields

    dumped = {}
    for field in fields(value):
        if not field.repr or not field.init:
            continue

        field_name = field.name
        field_value = getattr(value, field_name)

        if dumper.hide_defaults:
            if (default_value := field.default) is not MISSING:
                if field_value == default_value:
                    continue

            elif (default_factory := field.default_factory) is not MISSING:
                # Avoid calling the default factory function if its a known mutable type like list
                if default_factory in frozenset({list, set, dict}) and len(value) == 0:
                    continue

                # Avoid calling default factory if we know the value can't match the default
                cant_be_default = (
                    uuid := sys.modules.get("uuid")
                ) and default_factory in frozenset(
                    {
                        uuid.uuid1,
                        uuid.uuid4,
                        uuid.uuid6,
                        uuid.uuid7,
                    }
                )
                if not cant_be_default and field_value == default_factory():
                    continue

        alias = dumper._resolve_alias(type(value), type_path, field_name)
        dumped[alias] = dumper._dump(field_value, (*type_path, alias))

    return dumped


def dump_namedtuple(value: Any, type_path: TypePath, dumper: DumperProtocol) -> Dumped:
    dumped = {}
    for field_name in value._fields:
        field_value = getattr(value, field_name)

        if dumper.hide_defaults:
            if (
                field_default := value._field_defaults.get(field_name, _MISSING)
            ) is not _MISSING:
                if field_value == field_default:
                    continue

        alias = dumper._resolve_alias(type(value), type_path, field_name)
        dumped[alias] = dumper._dump(field_value, (*type_path, alias))
    return dumped


def dump_fspath(value: Any, type_path: TypePath, dumper: DumperProtocol) -> Dumped:
    return os.fspath(value)



__all__ = ["__version__"]

TYPE_CHECKING = False
if TYPE_CHECKING:
    __version__: str


def __getattr__(name: str) -> str:
    if name == "__version__":
        from importlib.metadata import version

        return version("tressed")
    raise AttributeError(f"Package '{__package__}' has no attribute '{name}'")



"""
Type predicates.

Do not import modules if not absolutely required. Instead, take advantage of the fact
that a type from a module can only be instantiated if it already has been imported.

For example to check if a type is an instance of foo.Bar, do not do:

    def is_foo_bar(type_form: TypeForm) -> bool:
        import foo

        try:
            return issubclass(t, foo.Bar)
        except TypeError:
            return False

But instead:

    def is_foo_bar(type_form: TypeForm) -> bool:
        if foo := sys.modules.get("foo"):
            try:
                return issubclass(t, foo.Bar)
            except TypeError:
                return False
        return False
"""

import sys

TYPE_CHECKING = False
if TYPE_CHECKING:
    from collections.abc import Callable

    from tressed.type_form import TypeForm

__all__ = [
    "get_origin",
    "get_args",
    "is_generic_tuple_type",
    "is_generic_homogeneous_tuple_type",
    "is_generic_list_type",
    "is_generic_set_type",
    "is_generic_frozenset_type",
    "is_dataclass_type",
    "is_newtype",
    "is_typeddict",
    "is_dict_type",
    "is_namedtuple_type",
    "is_uuid_type",
    "is_enum_type",
    "is_literal_type",
    "is_type_alias_type",
    "is_optional_type",
    "is_union_type",
    "is_fspath_type",
    "is_datetime_type",
    "is_discriminated_union",
]

if TYPE_CHECKING:
    type TypePredicate[T: TypeForm] = Callable[[T], bool]

    __all__ += [
        "TypeForm",
        "TypePredicate",
    ]


def get_origin(type_form: TypeForm) -> TypeForm | None:
    # Avoid importing typing module if possible.
    # TODO: Handle annotated, generic etc...
    return getattr(type_form, "__origin__", None)


def get_args(type_form: TypeForm) -> tuple[TypeForm, ...] | None:
    # Avoid importing typing module if possible.
    # TODO: Handle annotated, generic etc...
    return getattr(type_form, "__args__", None)


def is_generic_homogeneous_tuple_type(type_form: TypeForm) -> bool:
    """
    The given type form matches tuple[T, ...] or typing.Tuple[T, ...]
    """
    origin = get_origin(type_form)
    if origin is None:
        return False
    if origin is tuple or (
        (typing := sys.modules.get("typing")) and origin is typing.Tuple
    ):
        args = get_args(type_form)
        if args is None:
            return False
        return len(args) == 2 and args[1] == Ellipsis
    return False


def is_generic_tuple_type(type_form: TypeForm) -> bool:
    origin = get_origin(type_form)
    if origin is None:
        return False

    if get_args(type_form) is None:
        return False

    if origin is tuple:
        return True

    if typing := sys.modules.get("typing"):
        return origin is typing.Tuple

    return False


def is_generic_list_type(type_form: TypeForm) -> bool:
    origin = get_origin(type_form)
    if origin is None:
        return False

    if get_args(type_form) is None:
        return False

    if origin is list:
        return True

    if typing := sys.modules.get("typing"):
        return origin is typing.List

    return False


def is_generic_set_type(type_form: TypeForm) -> bool:
    origin = get_origin(type_form)
    if origin is None:
        return False

    if get_args(type_form) is None:
        return False

    if origin is set:
        return True

    if typing := sys.modules.get("typing"):
        return origin is typing.Set

    return False


def is_generic_frozenset_type(type_form: TypeForm) -> bool:
    origin = get_origin(type_form)
    if origin is None:
        return False

    if get_args(type_form) is None:
        return False

    if origin is frozenset:
        return True

    if typing := sys.modules.get("typing"):
        return origin is typing.FrozenSet

    return False


def is_dataclass_type(type_form: TypeForm) -> bool:
    if dataclasses := sys.modules.get("dataclasses"):
        # We want to match only dataclass types, not instances
        return dataclasses.is_dataclass(type_form) and isinstance(type_form, type)
    return False


if sys.version_info >= (3, 10):

    def is_newtype(type_form: TypeForm) -> bool:
        if typing := sys.modules.get("typing"):
            return type(type_form) is typing.NewType
        return False
else:
    # NewType used to be a function, the only way to identify it
    # was to check for the __supertype__ attribute.
    def is_newtype(type_form: TypeForm) -> bool:
        return hasattr(type_form, "__supertype__")


def is_ipaddress_type(type_form: TypeForm) -> bool:
    if ipaddress := sys.modules.get("ipaddress"):
        return type_form in {
            ipaddress.IPv4Address,
            ipaddress.IPv6Address,
            ipaddress.IPv4Interface,
            ipaddress.IPv6Interface,
            ipaddress.IPv4Network,
            ipaddress.IPv6Network,
        }
    return False


def is_typeddict(type_form: TypeForm) -> bool:
    if typing := sys.modules.get("typing"):
        if typing.is_typeddict(type_form):
            return True
    if typing_extensions := sys.modules.get("typing_extensions"):
        if typing_extensions.is_typeddict(type_form):
            return True
    return False


def is_dict_type(type_form: TypeForm) -> bool:
    origin = get_origin(type_form)
    if origin is dict:
        return True
    elif (typing := sys.modules.get("typing")) and origin is typing.Dict:
        return True
    return False


def is_namedtuple_type(type_form: TypeForm) -> bool:
    if (typing := sys.modules.get("typing")) and (
        orig_bases := getattr(type_form, "__orig_bases__", None)
    ):
        return typing.NamedTuple in orig_bases
    return False


def is_uuid_type(type_form: TypeForm) -> bool:
    if uuid := sys.modules.get("uuid"):
        return type_form is uuid.UUID
    return False


def is_enum_type(type_form: TypeForm) -> bool:
    if enum := sys.modules.get("enum"):
        return type(type_form) is enum.EnumType
    return False


def is_literal_type(type_form: TypeForm) -> bool:
    if typing := sys.modules.get("typing"):
        return get_origin(type_form) is typing.Literal
    return False


def is_type_alias_type(type_form: TypeForm) -> bool:
    return hasattr(type_form, "__type_params__") and hasattr(
        type_form, "evaluate_value"
    )


def is_optional_type(type_form: TypeForm) -> bool:
    """
    One of:
        T | None
        typing.Optional[T]
    """
    origin = get_origin(type_form)
    if origin is None:
        return False

    args = get_args(type_form)
    if args is None:
        return False

    num_args = len(args)
    if num_args == 2:
        if type(None) in args:
            import types

            if origin is types.UnionType:
                return True

            if typing := sys.modules.get("typing"):
                return origin is typing.Union

    elif num_args == 1 and (typing := sys.modules.get("typing")):
        return origin is typing.Optional

    return False


def is_union_type(type_form: TypeForm) -> bool:
    origin = get_origin(type_form)
    args = get_args(type_form)
    if not args:
        # A union needs at least one argument
        return False

    import types

    if origin is types.UnionType:
        return True

    if typing := sys.modules.get("typing"):
        return origin is typing.Union
    return False


def is_fspath_type(type_form: TypeForm) -> bool:
    return hasattr(type_form, "__fspath__")


def is_datetime_type(type_form: TypeForm) -> bool:
    if datetime := sys.modules.get("datetime"):
        return type_form in frozenset({datetime.datetime, datetime.date, datetime.time})
    return False


def is_discriminated_union(type_form: TypeForm) -> bool:
    """
    A discriminated union is a union annotated by one discriminator.
    """
    if not hasattr(type_form, "__metadata__"):
        return False

    if discriminated_union := sys.modules.get("tressed.discriminated_union"):
        args = get_args(type_form)
        assert args
        (arg,) = args
        if not is_union_type(arg):
            return False

        num_discriminators = sum(
            1
            for metadata in type_form.__metadata__
            if isinstance(metadata, discriminated_union.Discriminator)
        )
        return num_discriminators == 1
    return False



TYPE_CHECKING = False
if TYPE_CHECKING:
    from tressed.dumper.dumper import Dumper
    from tressed.loader.loader import Loader
    from tressed.type_form import TypeForm
    from tressed.type_path import TypePath


__all__ = ["Loader", "Dumper", "load", "dump", "TypeForm", "TypePath"]


if TYPE_CHECKING:
    _default_loader = Loader()
    load = _default_loader.load

    _default_dumper = Dumper()
    dump = _default_dumper.dump

else:

    def __getattr__(name: str):
        match name:
            case "Loader":
                from tressed.loader.loader import Loader

                return Loader

            case "Dumper":
                from tressed.dumper.dumper import Dumper

                return Dumper

            case "load":
                if "_default_loader" not in globals():
                    global _default_loader
                    from tressed.loader.loader import Loader

                    _default_loader = Loader()
                return _default_loader.load

            case "dump":
                if "_default_dumper" not in globals():
                    global _default_dumper
                    from tressed.dumper.dumper import Dumper

                    _default_dumper = Dumper()
                return _default_dumper.dump

            case "TypeForm":
                from tressed.type_form import TypeForm

                return TypeForm

            case "TypePath":
                from tressed.type_path import TypePath

                return TypePath

            case _:
                raise AttributeError(
                    f"Package '{__package__}' has no attribute '{name}'"
                )



import subprocess
import sys
from pathlib import Path


def test_lazy_imports(tmp_path: Path) -> None:
    """
    Make sure tressed uses lazy imports by default.
    Note that this is not fully reliable as pytest itself already loads a ton of modules.
    """
    prog = """\
import sys

def audit(event: str, args: tuple) -> None:
    match event:
        case "import":
            print(args[0])

sys.addaudithook(audit)

import tressed
"""
    prog_path = tmp_path / "prog.py"
    prog_path.write_text(prog)

    # Use audit hooks to detect imports, audit hooks cannot be removed so do it in a sub-process
    src_path = Path(__file__).parent.parent / "src"
    process = subprocess.run(
        [sys.executable, "-S", prog_path],
        check=True,
        capture_output=True,
        env={"PYTHONPATH": str(src_path)},
    )
    loaded_modules = set(process.stdout.decode("utf-8").splitlines(keepends=False))

    assert loaded_modules == {"tressed"}



TYPE_CHECKING = False
if TYPE_CHECKING:
    from collections.abc import Iterable
    from typing import Any, Final

    from tressed.loader.types import LoaderFn
    from tressed.type_form import TypeForm
    from tressed.type_path import TypePath, TypePathItem

__all__ = [
    "specialize_load_tuple",
    "specialize_load_simple_collection",
]


class Codegen:
    __slots__ = ("fn_name", "indent", "_parts", "_cached_code")

    def __init__(self, fn_name: str = "__specialized_fn") -> None:
        self.fn_name: Final = fn_name
        self.indent: Final = "    "

        self._parts: list[str] = []
        self._cached_code: str | None = None

    def code(self) -> str:
        if cached_code := self._cached_code:
            return cached_code
        import io

        builder = io.StringIO()
        builder.write(f"def {self.fn_name}(value, loader):\n")
        for line in self._parts:
            builder.write(line)
        builder.write("\n")
        cached_code = builder.getvalue()

        self._cached_code = cached_code
        self._parts.clear()
        return cached_code

    def exec(self) -> LoaderFn:
        # NOTE: This relies on the source being executed being only trusted and validated inputs.
        globals_: dict[str, Any] = {}
        locals_: dict[str, Any] = {}
        exec(self.code(), globals_, locals_)
        return locals_[self.fn_name]

    def _emit(self, part: str) -> None:
        self._parts.append(part)

    def _emit_indent(self, indent: int = 0) -> None:
        self._emit(self.indent * (indent + 1))

    def _emit_newline(self) -> None:
        self._emit("\n")

    def _emit_line(self, line: str, indent: int = 0) -> None:
        self._emit_indent(indent)
        self._emit(line)
        self._emit_newline()

    def emit_unpack_args(
        self, ident: str, args: tuple[TypeForm, ...]
    ) -> tuple[tuple[str, TypeForm], ...]:
        # TODO: Allocate identifiers!
        items = tuple((f"_item_{pos}", type_form) for pos, type_form in enumerate(args))
        match items:
            case []:
                pass
            case [(item_ident, _)]:
                self._emit_line(f"{item_ident}, = {ident}")
            case [*_]:
                self._emit_line(
                    f"{', '.join(item_ident for item_ident, _ in items)} = {ident}"
                )
        return items

    def emit_load_fn(self, ident: str = "_load") -> str:
        self._emit_line(f"{ident} = loader._load")
        return ident

    def emit_load(
        self,
        loader_ident: str,
        ident: str,
        type_form: TypeForm,
        type_path: TypePath,
        type_path_item: TypePathItem,
    ) -> str:
        # TODO: Allocate identifiers!
        loaded_ident = f"{ident}_loaded"
        # TODO: Handle type forms!
        self._emit_line(
            f"{loaded_ident} = _load({ident}, {type_form.__qualname__}, {_type_path_repr(*type_path, type_path_item)})"
        )
        return loaded_ident

    def emit_tuple(self, idents: Iterable[str], *, compact: bool = False) -> None:
        self._emit("(")
        if not compact:
            self._emit_newline()
        for ident in idents:
            if not compact:
                self._emit_indent(indent=1)
            self._emit(f"{ident},")
            if not compact:
                self._emit_newline()
        if not compact:
            self._emit_indent()
        self._emit(")")
        self._emit_newline()

    def emit_return(self) -> None:
        self._emit_indent()
        self._emit("return ")


def specialize_load_tuple[T](
    type_form: TypeForm[T],
    type_path: TypePath,
) -> str | None:
    """
    Generate specialized function for given type at given path.
    """
    from tressed.predicates import get_args

    args = get_args(type_form)
    assert args is not None

    codegen = Codegen()
    items = codegen.emit_unpack_args("value", args)
    load_fn = codegen.emit_load_fn()
    loaded = []
    for pos, (item, arg_type_form) in enumerate(items):
        loaded.append(codegen.emit_load(load_fn, item, arg_type_form, type_path, pos))
    codegen.emit_return()
    codegen.emit_tuple(loaded)
    return codegen.code()


def _type_form_repr(type_form) -> str:
    from tressed.predicates import get_args, get_origin

    origin = get_origin(type_form)
    if origin is not None:
        args = get_args(type_form)
        assert args is not None
        return _generic_type_repr(origin, args)

    return type_form.__qualname__


def _generic_type_repr(origin: TypeForm, args: tuple[TypeForm, ...]) -> str:
    if len(args) == 0:
        args_str = "()"
    else:
        args_str = ", ".join(map(_type_form_repr, args))
    return f"{origin.__qualname__}[{args_str}]"


class Ident:
    __slots__ = ("name",)

    def __init__(self, name: str) -> None:
        self.name = name


def _type_path_repr(*args: TypePathItem | Ident) -> str:
    items = []
    for arg in args:
        match arg:
            case Ident():
                items.append(arg.name)
            case _:
                items.append(repr(arg))

    match items:
        case [item]:
            # Special case for single item tuple
            return f"({item},)"
        case _:
            return f"({', '.join(items)})"


def specialize_load_simple_collection[T](
    type_form: TypeForm[T],
    type_path: TypePath,
) -> str | None:
    """
    Generate specialized function for given type at given path.
    """
    from tressed.predicates import get_args, get_origin

    origin = get_origin(type_form)
    assert origin is not None

    args = get_args(type_form)
    assert args is not None
    arg_type_form = args[0]

    codegen = Codegen()

    load_fn = codegen.emit_load_fn()
    codegen.emit_return()
    if origin is list:
        open, close = "[", "]"
    elif origin is set:
        open, close = "{", "}"
    elif origin is frozenset:
        open, close = "({", "})"
    else:
        open, close = f"{origin.__qualname__}([", "])"

    codegen._emit(f"""{open}
        {load_fn}(item, {_type_form_repr(arg_type_form)}, {_type_path_repr(*type_path, Ident("pos"))})
        for pos, item
        in enumerate(value)
    {close}
""")
    return codegen.code()



def test_to_camel() -> None:
    from tressed.alias import to_camel

    assert to_camel("foo_bar") == "fooBar"
    assert to_camel("foo") == "foo"
    assert to_camel("_foo") == "_foo"
    assert to_camel("_foO_bar__baz___") == "_fooBar__Baz___"
    assert to_camel("") == ""
    assert to_camel("_") == "_"
    assert to_camel("___") == "___"


def test_to_pascal() -> None:
    from tressed.alias import to_pascal

    assert to_pascal("foo_bar") == "FooBar"
    assert to_pascal("foo") == "Foo"
    assert to_pascal("_foo") == "_Foo"
    assert to_pascal("_foO_bar__baz___") == "_FooBar__Baz___"
    assert to_pascal("") == ""
    assert to_pascal("_") == "_"
    assert to_pascal("___") == "___"


def test_compose_alias_fn() -> None:
    from tressed.alias import compose_alias_fn

    def maybe_alias(name: str) -> str | None:
        match name:
            case "foo":
                return "FOO!"
            case _:
                return None

    def maybe_another_alias(name: str) -> str | None:
        match name:
            case "bar":
                return "bar?"
            case _:
                return None

    def definitely_an_alias(name: str) -> str:
        return name.upper()

    alias_fn = compose_alias_fn(
        maybe_alias,
        maybe_another_alias,
        definitely_an_alias,
    )

    assert alias_fn("foo", int, ()) == "FOO!"
    assert alias_fn("bar", int, ()) == "bar?"
    assert alias_fn("baz", int, ()) == "BAZ"
    assert alias_fn("buz", int, ()) == "BUZ"


def test_alias_resolver() -> None:
    from tressed.alias import AliasResolver

    # Simple alias
    alias_resolver = AliasResolver(alias_fn=lambda s: s.upper())
    assert alias_resolver.resolve("FooBar", int, ()) == "FOOBAR"

    # Type form alias
    alias_resolver = AliasResolver(
        alias_fn=lambda s, t: s.upper() if t is int else s.lower()
    )
    assert alias_resolver.resolve("FooBar", int, ()) == "FOOBAR"
    assert alias_resolver.resolve("FooBar", str, ()) == "foobar"

    # Type path alias
    alias_resolver = AliasResolver(alias_fn=lambda s, _, p: "/".join(map(str, (*p, s))))
    assert (
        alias_resolver.resolve("FooBar", int, ("foo", 1, "bar")) == "foo/1/bar/FooBar"
    )
    assert alias_resolver.resolve("FooBar", str, ()) == "FooBar"


def test_alias_resolver_caching_enabled() -> None:
    from tressed.alias import AliasResolver

    count = 0

    def count_alias(_: str) -> str:
        nonlocal count
        count += 1
        return str(count)

    alias_resolver = AliasResolver(alias_fn=count_alias, cache_resolved_aliases=True)
    assert alias_resolver.resolve("foo", int, ("abc", 1)) == "1"
    assert alias_resolver.resolve("foo", int, ("abc", 1)) == "1"
    assert alias_resolver.resolve("foo", str, ("abc", 1)) == "2"
    assert alias_resolver.resolve("foo", str, ("abc", 1)) == "2"
    assert alias_resolver.resolve("foo", str, ("abc", 2)) == "3"
    assert alias_resolver.resolve("foo", str, ("abc", 2)) == "3"


def test_alias_resolver_caching_disabled() -> None:
    from tressed.alias import AliasResolver

    count = 0

    def count_alias(_: str) -> str:
        nonlocal count
        count += 1
        return str(count)

    alias_resolver = AliasResolver(alias_fn=count_alias, cache_resolved_aliases=False)

    assert alias_resolver.resolve("foo", int, ("abc", 1)) == "1"
    assert alias_resolver.resolve("foo", int, ("abc", 1)) == "2"
    assert alias_resolver.resolve("foo", str, ("abc", 1)) == "3"
    assert alias_resolver.resolve("foo", str, ("abc", 1)) == "4"
    assert alias_resolver.resolve("foo", str, ("abc", 2)) == "5"
    assert alias_resolver.resolve("foo", str, ("abc", 2)) == "6"



__all__ = [
    "TypePathItem",
    "TypePath",
    "type_path_repr",
]

type TypePathItem = str | int
type TypePath = tuple[TypePathItem, ...]


def type_path_repr(type_path: TypePath) -> str:
    return f".{'.'.join(map(str, type_path))}"



import gc
import sys
from typing import NewType, assert_type

import pytest
from pytest_benchmark.fixture import BenchmarkFixture
from pytest_mock import MockerFixture

from tressed.exceptions import (
    TressedTypeFormError,
    TressedValueError,
)
from tressed.loader import Loader

TYPE_CHECKING = False
if TYPE_CHECKING:
    from typing import Any

    from tressed import TypeForm, TypePath
    from tressed.loader import LoaderProtocol


def test_load_identity() -> None:
    loader = Loader()
    assert loader.load(1, int) == 1
    assert loader.load(True, bool) is True
    assert loader.load(False, bool) is False
    assert loader.load("foo", str) == "foo"
    assert loader.load(None, type(None)) is None

    with pytest.raises(TressedValueError):
        loader.load(1, str)


def test_load_float() -> None:
    loader = Loader()

    assert loader.load(1.1, float) == 1.1
    assert loader.load(1, float) == 1.0


def test_load_complex() -> None:
    loader = Loader()

    loader.load([1.1, 2.2], complex) == 1.1 + 2.2j
    loader.load([1, 2.2], complex) == 1 + 2.2j
    loader.load([1.1, 2], complex) == 1.1 + 2j
    loader.load([1, 2], complex) == 1 + 2j


def test_load_list() -> None:
    loader = Loader()
    assert loader.load([1, 2, 3], list[int]) == [1, 2, 3]

    with pytest.raises(TressedTypeFormError):
        loader.load(["foo", "bar"], list)

    with pytest.raises(TressedValueError):
        loader.load(["foo", "bar"], list[int])


def test_load_list_legacy() -> None:
    from typing import List

    loader = Loader()
    assert loader.load([1, 2, 3], List[int]) == [1, 2, 3]

    with pytest.raises(TressedTypeFormError):
        loader.load(["foo", "bar"], List)

    with pytest.raises(TressedValueError):
        loader.load(["foo", "bar"], List[int])


def test_load_set() -> None:
    loader = Loader()
    loaded = loader.load([1, 2, 3, 2], set[int])
    assert type(loaded) is set
    assert loaded == {1, 2, 3}

    with pytest.raises(TressedTypeFormError):
        loader.load(["foo", "bar"], set)

    with pytest.raises(TressedValueError):
        loader.load(["foo", "bar"], set[int])


def test_load_set_legacy() -> None:
    from typing import Set

    loader = Loader()
    loaded = loader.load([1, 2, 3, 2], Set[int])
    assert type(loaded) is set
    assert loaded == {1, 2, 3}

    with pytest.raises(TressedTypeFormError):
        loader.load(["foo", "bar"], Set)

    with pytest.raises(TressedValueError):
        loader.load(["foo", "bar"], Set[int])


def test_load_frozenset() -> None:
    loader = Loader()
    loaded = loader.load([1, 2, 3, 2], frozenset[int])
    assert type(loaded) is frozenset
    assert loaded == frozenset({1, 2, 3})

    with pytest.raises(TressedTypeFormError):
        loader.load(["foo", "bar"], frozenset)

    with pytest.raises(TressedValueError):
        loader.load(["foo", "bar"], frozenset[int])


def test_load_frozenset_legacy() -> None:
    from typing import FrozenSet

    loader = Loader()
    loaded = loader.load([1, 2, 3, 2], FrozenSet[int])
    assert type(loaded) is frozenset
    assert loaded == frozenset({1, 2, 3})

    with pytest.raises(TressedTypeFormError):
        loader.load(["foo", "bar"], FrozenSet)

    with pytest.raises(TressedValueError):
        loader.load(["foo", "bar"], FrozenSet[int])


def test_load_homogeneous_tuple() -> None:
    loader = Loader()
    assert loader.load([1, 2, 3], tuple[int, ...]) == (1, 2, 3)

    with pytest.raises(TressedTypeFormError):
        loader.load(["foo", "bar"], tuple)

    with pytest.raises(TressedValueError):
        loader.load(["foo", "bar"], tuple[int, ...])


def test_load_homogeneous_tuple_legacy() -> None:
    from typing import Tuple

    loader = Loader()
    assert loader.load([1, 2, 3], Tuple[int, ...]) == (1, 2, 3)  # type: ignore[arg-type]

    with pytest.raises(TressedTypeFormError):
        loader.load(["foo", "bar"], Tuple)

    with pytest.raises(TressedValueError):
        loader.load(["foo", "bar"], Tuple[int, ...])  # type: ignore[arg-type]


def test_load_tuple() -> None:
    loader = Loader()
    assert loader.load([], tuple[()]) == ()
    assert loader.load([1], tuple[int]) == (1,)
    assert loader.load([1, True], tuple[int, bool]) == (1, True)
    assert loader.load([1, True, 1.1], tuple[int, bool, float]) == (1, True, 1.1)

    with pytest.raises(TressedValueError):
        loader.load([1, True, "foo"], tuple[int, bool, float])


def test_load_tuple_legacy() -> None:
    from typing import Tuple

    loader = Loader()
    assert loader.load([], Tuple[()]) == ()  # type: ignore[arg-type]
    assert loader.load([1], Tuple[int]) == (1,)  # type: ignore[arg-type]
    assert loader.load([1, True], Tuple[int, bool]) == (1, True)  # type: ignore[arg-type]
    assert loader.load([1, True, 1.1], Tuple[int, bool, float]) == (1, True, 1.1)  # type: ignore[arg-type]

    with pytest.raises(TressedValueError):
        loader.load([1, True, "foo"], Tuple[int, bool, float])  # type: ignore[arg-type]


def test_load_newtype() -> None:
    T = NewType("T", int)

    loader = Loader()
    loaded = loader.load(123, T)
    assert_type(loaded, T)

    assert loaded == T(123)


def test_load_ipaddress() -> None:
    import ipaddress

    loader = Loader()
    loaded_v4a = loader.load("127.0.0.1", ipaddress.IPv4Address)
    assert_type(loaded_v4a, ipaddress.IPv4Address)

    loaded_v6a = loader.load("::1", ipaddress.IPv6Address)
    assert_type(loaded_v6a, ipaddress.IPv6Address)

    loaded_v4n = loader.load("127.0.0.0/24", ipaddress.IPv4Network)
    assert_type(loaded_v4n, ipaddress.IPv4Network)

    loaded_v6n = loader.load("::0/64", ipaddress.IPv6Network)
    assert_type(loaded_v6n, ipaddress.IPv6Network)

    loaded_v4i = loader.load("127.0.0.1/24", ipaddress.IPv4Interface)
    assert_type(loaded_v4i, ipaddress.IPv4Interface)

    loaded_v6i = loader.load("::1/64", ipaddress.IPv6Interface)
    assert_type(loaded_v6i, ipaddress.IPv6Interface)


def test_load_tuple_benchmark(benchmark: BenchmarkFixture) -> None:
    loader = Loader()
    benchmark(loader.load, [1, True, 1.1], tuple[int, bool, float])


def test_load_tuple_specialized_benchmark(benchmark: BenchmarkFixture) -> None:
    loader = Loader(enable_specialization=True)
    benchmark(loader.load, [1, True, 1.1], tuple[int, bool, float])


def test_load_set_benchmark(benchmark: BenchmarkFixture) -> None:
    loader = Loader()
    benchmark(loader.load, [[1, "two", 3.3]] * 30, set[tuple[int, str, float]])


def test_load_set_specialized_benchmark(benchmark: BenchmarkFixture) -> None:
    loader = Loader(enable_specialization=True)
    benchmark(loader.load, [[1, "two", 3.3]] * 30, set[tuple[int, str, float]])


def test_load_unknown_type() -> None:
    # Loading an unknown type will go through all the predicates,
    # make sure the predicates do not load any additional modules.
    if "dataclasses" in sys.modules:
        del sys.modules["dataclasses"]
        gc.collect()

    modules_before = frozenset(sys.modules.keys())
    assert "dataclasses" not in modules_before

    loader = Loader()

    class UnknownType:
        def __repr__(self) -> str:
            return f"{self.__class__.__name__}()"

    with pytest.raises(TressedTypeFormError) as exc_info:
        loader.load(1, UnknownType)
    assert (
        str(exc_info.value) == "Unhandled type form UnknownType at path . for value 1"
    )

    modules_after = frozenset(sys.modules.keys())
    new_modules = frozenset(
        {
            module
            for module in modules_after - modules_after
            # Ignore internal modules
            if not module.startswith("tressed.")
        }
    )
    assert new_modules == frozenset()


def test_load_custom_type() -> None:
    class CustomType:
        def __init__(self, value: int) -> None:
            self.value = value

        def __repr__(self) -> str:
            return f"{self.__class__.__name__}(value={self.value})"

        def __eq__(self, other: Any) -> bool:
            return type(other) is type(self) and other.value == self.value

    def _load_custom_type_double(
        value: Any, type_form: TypeForm, type_path: TypePath, loader: LoaderProtocol
    ) -> Any:
        return CustomType(int(value) * 2)

    loader = Loader(extra_type_loaders={CustomType: _load_custom_type_double})
    assert loader.load("123", CustomType) == CustomType(value=246)

    def _is_custom_type(type_form: TypeForm) -> bool:
        return type(type_form) is type and issubclass(type_form, CustomType)

    def _load_custom_type_third(
        value: Any, type_form: TypeForm, type_path: TypePath, loader: LoaderProtocol
    ) -> Any:
        return CustomType(int(value / 3))

    loader = Loader(extra_type_mappers={_is_custom_type: _load_custom_type_third})
    assert loader.load(123.4, CustomType) == CustomType(value=41)


def test_barebones_loader() -> None:
    type UpperStr = str
    type LowerStr = str

    def _load_float_times_five(
        value: Any, type_form: TypeForm, type_path: TypePath, loader: LoaderProtocol
    ) -> Any:
        if type(value) is not float:
            raise TressedValueError(
                value,
                type_form,
                type_path,
                message="Sorry pal, can't accept an int for a float!",
            )
        return value * 5

    loader = Loader(
        default_type_loaders={float: _load_float_times_five},
        default_type_mappers={
            lambda tf: tf in (UpperStr, LowerStr): lambda v, tf, __, ___: v.upper()
            if tf is UpperStr
            else v.lower(),
        },
    )

    # Allow floats, but multiply their values by five, don't allow passing an int as float!
    assert loader.load(5.0, float) == 25.0
    with pytest.raises(TressedValueError) as exc_info_2:
        assert loader.load(5, float) == 25
    assert (
        str(exc_info_2.value)
        == "Failed to load value of type int at path . into type form float: Sorry pal, can't accept an int for a float!"
    )

    # Don't allow regular str, only UpperStr and LowerStr
    with pytest.raises(TressedTypeFormError) as exc_info_1:
        loader.load("FooBar", str)
    assert (
        str(exc_info_1.value) == "Unhandled type form str at path . for value 'FooBar'"
    )
    assert loader.load("FooBar", UpperStr) == "FOOBAR"
    assert loader.load("FooBar", LowerStr) == "foobar"


def test_load_dataclass(mocker: MockerFixture) -> None:
    from dataclasses import dataclass, field

    @dataclass
    class SomeDataclass:
        foo: str
        bar: str = "bar"
        baz: list[int] = field(default_factory=lambda: [1, 2, 3])
        bar_bar: tuple[int, str] = field(metadata={"alias": "barBar"}, kw_only=True)

    value = {
        "foo": "foo",
        "barBar": (2, "humbug"),
    }
    expected = SomeDataclass(
        foo="foo",
        bar="bar",
        baz=[1, 2, 3],
        bar_bar=(2, "humbug"),
    )
    loader = Loader()
    assert loader.load(value, SomeDataclass) == expected

    assert loader._alias_resolver._cache == {
        ("foo", SomeDataclass, ()): "foo",
        ("bar", SomeDataclass, ()): "bar",
        ("baz", SomeDataclass, ()): "baz",
        ("bar_bar", SomeDataclass, ()): "barBar",
    }


def test_load_dataclass_caching_enabled(mocker: MockerFixture) -> None:
    from dataclasses import dataclass, field

    @dataclass
    class SomeDataclass:
        foo: str
        bar: str = "bar"
        baz: list[int] = field(default_factory=lambda: [1, 2, 3])
        bar_bar: tuple[int, str] = field(metadata={"alias": "barBar"}, kw_only=True)

    count = 0

    def _count_alias_fn(name: str) -> str:
        nonlocal count
        count += 1
        return name

    value = {
        "foo": "foo",
        "barBar": (2, "humbug"),
    }
    expected = SomeDataclass(
        foo="foo",
        bar="bar",
        baz=[1, 2, 3],
        bar_bar=(2, "humbug"),
    )
    loader = Loader(alias_fn=_count_alias_fn)
    assert loader.load(value, SomeDataclass) == expected
    assert count == 3

    assert loader.load(value, SomeDataclass) == expected
    assert count == 3


def test_load_dataclass_caching_disabled(mocker: MockerFixture) -> None:
    from dataclasses import dataclass, field

    from tressed.alias import AliasResolver

    @dataclass
    class SomeDataclass:
        foo: str
        bar: str = "bar"
        baz: list[int] = field(default_factory=lambda: [1, 2, 3])
        bar_bar: tuple[int, str] = field(metadata={"alias": "barBar"}, kw_only=True)

    count = 0

    def _count_alias_fn(name: str) -> str:
        nonlocal count
        count += 1
        return name

    value = {
        "foo": "foo",
        "barBar": (2, "humbug"),
    }
    expected = SomeDataclass(
        foo="foo",
        bar="bar",
        baz=[1, 2, 3],
        bar_bar=(2, "humbug"),
    )
    loader = Loader(
        alias_fn=_count_alias_fn,
        alias_resolver_factory=lambda alias_fn: AliasResolver(
            alias_fn, cache_resolved_aliases=False
        ),
    )
    assert loader.load(value, SomeDataclass) == expected
    assert count == 3
    assert loader._alias_resolver._cache == {}

    assert loader.load(value, SomeDataclass) == expected
    assert count == 6
    assert loader._alias_resolver._cache == {}


def test_load_dataclass_alias_fn(mocker: MockerFixture) -> None:
    from dataclasses import dataclass, field

    from tressed.alias import to_camel

    @dataclass
    class SomeDataclass:
        foo_foo: str
        baz_baz: list[int] = field(default_factory=lambda: [1, 2, 3])
        bar_bar: tuple[int, str] = field(metadata={"alias": "BAR_BAR"}, kw_only=True)

    value = {
        "fooFoo": "foo",
        "BAR_BAR": (2, "humbug"),
    }
    expected = SomeDataclass(
        foo_foo="foo",
        baz_baz=[1, 2, 3],
        bar_bar=(2, "humbug"),
    )
    loader = Loader(alias_fn=to_camel)
    assert loader.load(value, SomeDataclass) == expected
    assert loader._alias_resolver._cache == {
        ("foo_foo", SomeDataclass, ()): "fooFoo",
        ("baz_baz", SomeDataclass, ()): "bazBaz",
        ("bar_bar", SomeDataclass, ()): "BAR_BAR",
    }


def test_load_dict() -> None:
    loader = Loader()

    loaded = loader.load({"foo": [1, 2], "bar": [5, 6, 6, 8]}, dict[str, set[int]])
    assert_type(loaded, dict[str, set[int]])
    assert loaded == {
        "foo": {1, 2},
        "bar": {5, 6, 8},
    }


def test_load_legacy_dict() -> None:
    from typing import Dict

    loader = Loader()

    loaded = loader.load({"foo": [1, 2], "bar": [5, 6, 6, 8]}, Dict[str, set[int]])
    assert_type(loaded, dict[str, set[int]])
    assert loaded == {
        "foo": {1, 2},
        "bar": {5, 6, 8},
    }


def test_load_typeddict() -> None:
    from typing import TypedDict

    loader = Loader()

    class SomeTypedDict(TypedDict):
        foo: int
        bar: str

    loaded = loader.load({"foo": 123, "bar": "BAR"}, SomeTypedDict)
    assert_type(loaded, SomeTypedDict)
    assert loaded == {"foo": 123, "bar": "BAR"}


def test_load_typeddict_optional_keys() -> None:
    from typing import TypedDict

    loader = Loader()

    class SomeTypedDict(TypedDict, total=False):
        foo: int
        bar: str

    loaded = loader.load({"foo": 123}, SomeTypedDict)
    assert_type(loaded, SomeTypedDict)
    assert loaded == {"foo": 123}

    loaded = loader.load({"foo": 123, "bar": "BAR", "baz": "BAZ"}, SomeTypedDict)
    assert_type(loaded, SomeTypedDict)
    assert loaded == {"foo": 123, "bar": "BAR", "baz": "BAZ"}


def test_load_typeddict_closed() -> None:
    from typing_extensions import TypedDict

    loader = Loader()

    class SomeTypedDict(TypedDict, closed=True):  # type: ignore[call-arg]
        foo: int

    with pytest.raises(TressedValueError) as exc_info:
        loader.load(
            {"foo": 123, "bar": "bar", "buz": "BUZ", "baz": "BAZ"}, SomeTypedDict
        )

    assert str(exc_info.value) == (
        "Failed to load value of type dict at path . into type form SomeTypedDict: "
        "extra keys 'bar', 'baz', 'buz': {'foo': 123, 'bar': 'bar', 'buz': 'BUZ', 'baz': 'BAZ'}"
    )


def test_load_typeddict_extra_items() -> None:
    from typing_extensions import TypedDict

    loader = Loader()

    class SomeTypedDict(TypedDict, extra_items=str):  # type: ignore[call-arg]
        foo: int
        bar: str
        baz: str

    loaded = loader.load(
        {"foo": 123, "bar": "bar", "buz": "BUZ", "baz": "BAZ"}, SomeTypedDict
    )
    assert_type(loaded, SomeTypedDict)
    assert loaded == {"foo": 123, "bar": "bar", "buz": "BUZ", "baz": "BAZ"}


def test_load_namedtuple() -> None:
    from typing import NamedTuple

    class SomeNamedTuple(NamedTuple):
        foo: tuple[int, float]
        bar: str = "bar"

    loader = Loader()
    assert loader.load(
        {"foo": [1, 1.1], "bar": "BAR!"}, SomeNamedTuple
    ) == SomeNamedTuple(foo=(1, 1.1), bar="BAR!")
    assert loader.load({"foo": [1, 1.1]}, SomeNamedTuple) == SomeNamedTuple(
        foo=(1, 1.1), bar="bar"
    )

    with pytest.raises(TressedValueError):
        assert loader.load(
            {"foo": [1, 1.1], "baz": "baz"}, SomeNamedTuple
        ) == SomeNamedTuple(foo=(1, 1.1), bar="bar")
    with pytest.raises(TressedValueError):
        assert loader.load({"bar": "BAR!"}, SomeNamedTuple) == SomeNamedTuple(
            foo=(1, 1.1), bar="BAR!"
        )


def test_load_from_argparse_namespace() -> None:
    from argparse import Namespace
    from typing import NamedTuple

    class SomeNamedTuple(NamedTuple):
        foo: tuple[int, float]
        bar: str = "bar"

    loader = Loader()
    assert loader.load(
        Namespace(foo=[1, 1.1], bar="BAR!"), SomeNamedTuple
    ) == SomeNamedTuple(foo=(1, 1.1), bar="BAR!")


def test_load_enum() -> None:
    from enum import Enum, IntEnum, StrEnum, auto

    class SomeEnum(Enum):
        FOO = "foo"
        BAR = "bar"

    loader = Loader()
    assert loader.load("foo", SomeEnum) == SomeEnum.FOO

    class SomeStrEnum(StrEnum):
        FOO = auto()
        BAR = auto()

    assert loader.load("bar", SomeStrEnum) == SomeStrEnum.BAR

    class SomeIntEnum(IntEnum):
        FOO = auto()
        BAR = auto()

    assert loader.load(1, SomeIntEnum) == SomeIntEnum.FOO


def test_load_uuid() -> None:
    import uuid

    loader = Loader()

    # uuid1
    assert loader.load("417c0ae2-d114-11f0-b300-d7c04d362c1f", uuid.UUID) == uuid.UUID(
        "417c0ae2-d114-11f0-b300-d7c04d362c1f"
    )
    # uuid4
    assert loader.load("c932f581-6430-4ae1-ad63-85489a0206b2", uuid.UUID) == uuid.UUID(
        "c932f581-6430-4ae1-ad63-85489a0206b2"
    )
    # uuid6
    assert loader.load("1f0d1144-17c1-6152-a18b-98af65b3c73e", uuid.UUID) == uuid.UUID(
        "1f0d1144-17c1-6152-a18b-98af65b3c73e"
    )
    # uuid7
    assert loader.load("019ae987-1dc6-770d-88c2-ec9b62f79444", uuid.UUID) == uuid.UUID(
        "019ae987-1dc6-770d-88c2-ec9b62f79444"
    )


def test_load_literal() -> None:
    from typing import Literal

    loader = Loader()

    Foo = Literal[1, 2, 3, "foo"]
    assert loader.load(2, Foo) == 2
    assert loader.load("foo", Foo) == "foo"
    with pytest.raises(TressedValueError) as exc_info:
        loader.load(5, Foo)
    assert str(exc_info.value) == (
        "Failed to load value of type int at path . into type form Literal[1, 2, 3, 'foo']: "
        "got value 5 but expected one of: 1, 2, 3, 'foo'"
    )


def test_load_type_alias() -> None:
    from typing import Literal

    loader = Loader()

    type TruthyLiteral = Literal[True, "true", "t", "yes", "y", "1"]
    type IntPair = tuple[int, int]
    assert loader.load("yes", TruthyLiteral) == "yes"
    assert loader.load([1, 2], IntPair) == (1, 2)


def test_load_generic_type_alias() -> None:
    loader = Loader()

    type Pair[T] = tuple[T, T]
    type IntPair = Pair[int]
    type StrPair = Pair[str]

    assert loader.load((1, 0), IntPair) == (1, 0)
    assert loader.load(("yes", "no"), StrPair) == ("yes", "no")

    with pytest.raises(TressedValueError) as exc_info:
        loader.load((1, 0), Pair) == (1, 0)
    assert str(exc_info.value) == (
        "Failed to load value of type tuple at path . into type form Pair[T=?]: "
        "type form should have only concrete type parameters"
    )

    type SomeMapping[K, V] = dict[K, V]
    type IntMapping[K] = dict[K, int]

    assert loader.load({"foo": 1}, IntMapping[str]) == {"foo": 1}

    with pytest.raises(TressedValueError) as exc_info:
        loader.load({"foo": 1}, IntMapping)
    assert str(exc_info.value) == (
        "Failed to load value of type dict at path . into type form IntMapping[K=?]: "
        "type form should have only concrete type parameters"
    )

    with pytest.raises(TressedValueError) as exc_info:
        loader.load({"foo": 1}, SomeMapping[int])  # type: ignore[arg-type]
    assert str(exc_info.value) == (
        "Failed to load value of type dict at path . into type form SomeMapping[K=int, V=?]: "
        "type form should have only concrete type parameters"
    )


def test_load_optional() -> None:
    loader = Loader()

    # T | None
    assert loader.load(1, int | None) == 1
    assert loader.load(1, None | int) == 1
    assert loader.load(None, int | None) is None
    assert loader.load(None, None | int) is None
    with pytest.raises(TressedValueError):
        assert loader.load("foo", int | None)


def test_load_optional_legacy() -> None:
    from typing import Optional, Union

    loader = Loader()

    # typing.Optional[T]
    assert loader.load(1, Optional[int]) == 1
    assert loader.load(None, Optional[int]) is None
    assert loader.load(None, Optional[None]) is None
    with pytest.raises(TressedValueError):
        assert loader.load("foo", Optional[int])

    # typing.Union[T, None], typing.Union[None, T]
    assert loader.load(1, Union[int, None]) == 1
    assert loader.load(None, Union[int, None]) is None
    assert loader.load(1, Union[None, int]) == 1
    assert loader.load(None, Union[None, int]) is None
    assert loader.load(None, Union[None, None]) is None
    with pytest.raises(TressedValueError):
        assert loader.load("foo", Union[int, None])


def test_load_union() -> None:
    loader = Loader()

    assert loader.load(1, int | str) == 1
    assert loader.load("foo", int | str) == "foo"

    with pytest.raises(TressedValueError) as exc_info:
        loader.load([1, 2], int | str)
    assert (
        str(exc_info.value)
        == "Failed to load value of type list at path . into type form int | str (2 sub-exceptions)"
    )
    assert [str(e) for e in exc_info.value.exceptions] == [
        "Failed to load value of type list at path . into type form int",
        "Failed to load value of type list at path . into type form str",
    ]


def test_load_legacy_union() -> None:
    from typing import Union

    loader = Loader()

    assert loader.load(1, Union[int, str]) == 1
    assert loader.load("foo", Union[int, str]) == "foo"

    with pytest.raises(TressedValueError) as exc_info:
        loader.load([1, 2], Union[int, str])
    assert (
        str(exc_info.value)
        == "Failed to load value of type list at path . into type form int | str (2 sub-exceptions)"
    )
    assert [str(e) for e in exc_info.value.exceptions] == [
        "Failed to load value of type list at path . into type form int",
        "Failed to load value of type list at path . into type form str",
    ]


def test_load_path() -> None:
    from pathlib import Path, PurePosixPath, PureWindowsPath

    loader = Loader()

    assert loader.load("/foo/bar/baz", Path) == Path("/foo/bar/baz")
    assert loader.load("/foo/bar/baz", PurePosixPath) == PurePosixPath("/foo/bar/baz")
    assert loader.load("/foo/bar/baz", PureWindowsPath) == PureWindowsPath(
        "/foo/bar/baz"
    )


def test_load_datetime() -> None:
    from datetime import UTC, date, datetime, time, timedelta, timezone

    loader = Loader()
    assert loader.load("12:45:59", time) == time(12, 45, 59)
    assert loader.load("12:45:59Z", time) == time(12, 45, 59, tzinfo=UTC)
    assert loader.load("12:45:59-00:30", time) == time(
        12, 45, 59, tzinfo=timezone(timedelta(seconds=-1800))
    )
    assert loader.load("2025-12-29", date) == date(2025, 12, 29)
    assert loader.load("2025-12-29T12:45:59", datetime) == datetime(
        2025, 12, 29, 12, 45, 59
    )
    assert loader.load("2025-12-29T12:45:59Z", datetime) == datetime(
        2025, 12, 29, 12, 45, 59, tzinfo=UTC
    )
    assert loader.load("2025-12-29T12:45:59+02:00", datetime) == datetime(
        2025, 12, 29, 12, 45, 59, tzinfo=timezone(timedelta(seconds=7200))
    )


def test_load_discriminated_union_first_match() -> None:
    from typing import Annotated, Literal, NamedTuple, get_args, get_type_hints

    from tressed.discriminated_union import Discriminator

    class Foo(NamedTuple):
        tag: Literal["foo"]
        field1: str
        field2: int
        field4: tuple[int, int] = (6, 6)

    class Bar(NamedTuple):
        tag: Literal["bar", "BAR"]
        field2: int
        field3: float
        field5: tuple[int, int] = (6, 7)

    loader = Loader()

    def _match_tag(value: Any, type_form: TypeForm) -> bool:
        tags = get_args(get_type_hints(type_form)["tag"])
        return value["tag"] in tags

    type TaggedUnion = Annotated[Foo | Bar, Discriminator(_match_tag)]

    assert loader.load(
        {"tag": "foo", "field1": "hej", "field2": 123}, TaggedUnion
    ) == Foo(tag="foo", field1="hej", field2=123, field4=(6, 6))
    assert loader.load(
        {"tag": "bar", "field2": 123, "field3": 1.23}, TaggedUnion
    ) == Bar(tag="bar", field2=123, field3=1.23, field5=(6, 7))
    assert loader.load(
        {"tag": "BAR", "field2": 123, "field3": 1.23}, TaggedUnion
    ) == Bar(tag="BAR", field2=123, field3=1.23, field5=(6, 7))
    with pytest.raises(TressedValueError) as exc_info:
        loader.load({"tag": "baz", "field2": 123, "field3": 1.23}, TaggedUnion)
    assert str(exc_info.value) == (
        "Failed to load value of type dict at path . into type form Annotated[Foo | Bar]: "
        "value did not match discriminated union discriminant"
    )


def test_load_discriminated_union_best_match() -> None:
    from typing import Annotated, NamedTuple

    from tressed.discriminated_union import Discriminator

    class Foo(NamedTuple):
        field1: str
        field2: int
        field4: tuple[int, int] = (6, 6)

    class Bar(NamedTuple):
        field2: int
        field3: float
        field5: tuple[int, int] = (6, 7)

    loader = Loader()

    def _match_num_fields(value: Any, type_form: TypeForm) -> int:
        print(f"{value=} {type_form=} {set(type_form._fields) & set(value.keys())=}")
        return len(set(type_form._fields) & set(value.keys()))

    type BestUnion = Annotated[
        Foo | Bar, Discriminator(_match_num_fields, strategy="best-match")
    ]

    assert loader.load({"field1": "foo", "field2": 123}, BestUnion) == Foo(
        field1="foo", field2=123, field4=(6, 6)
    )
    assert loader.load({"field2": 123, "field3": 6.7}, BestUnion) == Bar(
        field2=123, field3=6.7, field5=(6, 7)
    )

    with pytest.raises(TressedValueError) as exc_info:
        loader.load({"foo": "bar"}, BestUnion)
    assert str(exc_info.value) == (
        "Failed to load value of type dict at path . into type form Annotated[Foo | Bar]: "
        "value did not match discriminated union discriminant"
    )

    # ambiguous
    with pytest.raises(TressedValueError) as exc_info:
        loader.load({"field2": "bar"}, BestUnion)
    assert str(exc_info.value) == (
        "Failed to load value of type dict at path . into type form Annotated[Foo | Bar]: "
        "value did not match discriminated union discriminant"
    )



"""
Load the specified value from python basic types suitable for serialization to the given type form.

Supports aliases.
"""

from __future__ import annotations

from tressed.exceptions import TressedTypeFormError, TressedValueError

TYPE_CHECKING = False
if TYPE_CHECKING:
    from collections.abc import Callable, Mapping
    from typing import Any

    from tressed.alias import Alias, AliasFn, AliasResolver
    from tressed.loader.types import LoaderFn, TypePath
    from tressed.predicates import TypePredicate
    from tressed.type_form import TypeForm


__all__ = [
    "Loader",
]


def _default_type_loaders() -> dict[TypeForm, LoaderFn]:
    from tressed.loader.loaders import load_complex, load_float, load_identity

    return {
        bool: load_identity,
        int: load_identity,
        float: load_float,
        str: load_identity,
        type(None): load_identity,
        complex: load_complex,
    }


def _default_type_mappers(specialize: bool) -> dict[TypePredicate, LoaderFn]:
    from tressed.loader.loaders import (
        load_dataclass,
        load_datetime,
        load_dict,
        load_discriminated_union,
        load_literal,
        load_namedtuple,
        load_newtype,
        load_optional,
        load_simple_collection,
        load_simple_scalar,
        load_tuple,
        load_type_alias,
        load_typeddict,
        load_union,
    )
    from tressed.predicates import (
        is_dataclass_type,
        is_datetime_type,
        is_dict_type,
        is_discriminated_union,
        is_enum_type,
        is_fspath_type,
        is_generic_frozenset_type,
        is_generic_homogeneous_tuple_type,
        is_generic_list_type,
        is_generic_set_type,
        is_generic_tuple_type,
        is_ipaddress_type,
        is_literal_type,
        is_namedtuple_type,
        is_newtype,
        is_optional_type,
        is_type_alias_type,
        is_typeddict,
        is_union_type,
        is_uuid_type,
    )

    load_tuple_ = load_tuple
    load_simple_collection_ = load_simple_collection
    if specialize:
        from tressed.loader.specializer import SpecializingLoader
        from tressed.loader.specializers import (
            specialize_load_simple_collection,
            specialize_load_tuple,
        )

        load_tuple_ = SpecializingLoader(load_tuple, specialize_load_tuple)
        load_simple_collection_ = SpecializingLoader(
            load_simple_collection, specialize_load_simple_collection
        )

    # Note that the order matters as some predicates match several types,
    # put the most specific match first.
    return {
        is_generic_homogeneous_tuple_type: load_simple_collection_,
        is_generic_tuple_type: load_tuple_,
        is_generic_list_type: load_simple_collection_,
        is_generic_set_type: load_simple_collection_,
        is_generic_frozenset_type: load_simple_collection_,
        is_dict_type: load_dict,
        is_literal_type: load_literal,
        is_type_alias_type: load_type_alias,
        is_optional_type: load_optional,
        # NOTE: Union has to be after optional, since optionals of the form T | None are also unions.
        is_union_type: load_union,
        is_discriminated_union: load_discriminated_union,
        is_newtype: load_newtype,
        is_typeddict: load_typeddict,
        is_dataclass_type: load_dataclass,
        is_ipaddress_type: load_simple_scalar,
        is_namedtuple_type: load_namedtuple,
        is_enum_type: load_simple_scalar,
        is_uuid_type: load_simple_scalar,
        is_fspath_type: load_simple_scalar,
        is_datetime_type: load_datetime,
    }


class Loader:
    def __init__(
        self,
        *,
        default_type_loaders: Mapping[TypeForm, LoaderFn] | None = None,
        default_type_mappers: Mapping[TypePredicate, LoaderFn] | None = None,
        extra_type_loaders: Mapping[TypeForm, LoaderFn] | None = None,
        extra_type_mappers: Mapping[TypePredicate, LoaderFn] | None = None,
        enable_specialization: bool = False,
        # If set enables alias lookup on fields, for example for dataclasses.
        alias_field: str | None = "alias",
        # Pass custom alias function, for example mapping field names to camelCase using to_camel.
        alias_fn: AliasFn | None = None,
        # Overriding the resolver factory allows for more advanced behavior like disablign caching,
        # or changing the alias resolution behavior entirely.
        alias_resolver_factory: Callable[[AliasFn | None], AliasResolver] | None = None,
    ) -> None:
        # Map a type form to its loader
        if default_type_loaders is None:
            type_loaders = _default_type_loaders()
        else:
            type_loaders = dict(default_type_loaders)
        if extra_type_loaders:
            type_loaders |= extra_type_loaders
        self._type_loaders: dict[TypeForm, LoaderFn] = type_loaders

        # Mapping of type predicate to a loader
        if default_type_mappers is None:
            type_mappers = _default_type_mappers(enable_specialization)
        else:
            type_mappers = dict(default_type_mappers)
        if extra_type_mappers:
            type_mappers |= extra_type_mappers
        self._type_mappers: Mapping[TypePredicate, LoaderFn] = type_mappers

        from tressed.alias import (
            AliasResolver,
            compose_alias_fn,
            identity_alias_fn,
            make_maybe_dataclass_alias_fn,
        )

        if alias_field:
            alias_fn = compose_alias_fn(
                make_maybe_dataclass_alias_fn(alias_field=alias_field),
                alias_fn if alias_fn is not None else identity_alias_fn,
            )

        if alias_resolver_factory is None:
            self._alias_resolver = AliasResolver(alias_fn)
        else:
            self._alias_resolver = alias_resolver_factory(alias_fn)

    def _resolve_alias[T](
        self, type_form: TypeForm, type_path: TypePath, name: str
    ) -> Alias:
        return self._alias_resolver.resolve(name, type_form, type_path)

    def _load[T](self, value: Any, type_form: TypeForm[T], type_path: TypePath) -> T:
        if (type_loader := self._type_loaders.get(type_form)) is None:
            for type_predicate, type_loader in self._type_mappers.items():
                if type_predicate(type_form):
                    break
            else:
                type_loader = None
            if type_loader is None:
                raise TressedTypeFormError(value, type_form, type_path)

            # Cache lookup for next time
            self._type_loaders[type_form] = type_loader

        try:
            return type_loader(value, type_form, type_path, self)
        except TressedValueError:
            raise
        except Exception as e:
            error = TressedValueError(value, type_form, type_path)
            error.add_note(f"{type(e)}: {e}")
            raise error from e

    def load[T](self, value: Any, type_form: TypeForm[T]) -> T:
        return self._load(value, type_form, ())



def test_specialize_load_tuple() -> None:
    from tressed.loader import Loader
    from tressed.loader.specializers import specialize_load_tuple

    code = specialize_load_tuple(tuple[int, float, str], ("foo", 1))
    assert (
        code
        == """\
def __specialized_fn(value, loader):
    _item_0, _item_1, _item_2 = value
    _load = loader._load
    _item_0_loaded = _load(_item_0, int, ('foo', 1, 0))
    _item_1_loaded = _load(_item_1, float, ('foo', 1, 1))
    _item_2_loaded = _load(_item_2, str, ('foo', 1, 2))
    return (
        _item_0_loaded,
        _item_1_loaded,
        _item_2_loaded,
    )

"""
    )
    loader = Loader()
    globals_: dict = {}
    locals_: dict = {}
    exec(code, globals_, locals_)
    specialized_fn = locals_["__specialized_fn"]
    assert specialized_fn([1, 1.1, "foobar"], loader) == (1, 1.1, "foobar")


def test_specialize_load_simple_collection() -> None:
    from tressed.loader import Loader
    from tressed.loader.specializers import specialize_load_simple_collection

    code = specialize_load_simple_collection(
        set[tuple[int, str, float]], ("foo", 0, "bar")
    )
    assert (
        code
        == """\
def __specialized_fn(value, loader):
    _load = loader._load
    return {
        _load(item, tuple[int, str, float], ('foo', 0, 'bar', pos))
        for pos, item
        in enumerate(value)
    }

"""
    )
    loader = Loader()
    globals_: dict = {}
    locals_: dict = {}
    exec(code, globals_, locals_)
    specialized_fn = locals_["__specialized_fn"]
    assert specialized_fn(
        [[1, "two", 3.3], [4, "five", 5.5], [1, "two", 3.3]], loader
    ) == {(1, "two", 3.3), (4, "five", 5.5)}



"""
Dump the specified value to python basic types suitable for serialization.
Note that contrarily from Loader, the Dumper operates on concrete values and types.

Supports aliases.
"""

from tressed.exceptions import TressedTypeError, TressedValueError

TYPE_CHECKING = False
if TYPE_CHECKING:
    from collections.abc import Callable, Mapping
    from typing import Any

    from tressed.alias import Alias, AliasFn, AliasResolver
    from tressed.dumper.types import Dumped, DumperFn
    from tressed.predicates import TypePredicate
    from tressed.type_path import TypePath

__all__ = ["Dumper"]


def _default_type_dumpers() -> dict[type, DumperFn]:
    from tressed.dumper.dumpers import (
        dump_complex,
        dump_identity,
        dump_simple_mapping,
        dump_simple_sequence,
    )

    return {
        str: dump_identity,
        int: dump_identity,
        float: dump_identity,
        bool: dump_identity,
        type(None): dump_identity,
        list: dump_simple_sequence,
        tuple: dump_simple_sequence,
        set: dump_simple_sequence,
        frozenset: dump_simple_sequence,
        dict: dump_simple_mapping,
        complex: dump_complex,
    }


def _default_type_mappers(specialize: bool) -> dict[TypePredicate, DumperFn]:
    from tressed.dumper.dumpers import (
        dump_dataclass,
        dump_datetime,
        dump_enum,
        dump_fspath,
        dump_namedtuple,
        dump_simple_scalar,
    )
    from tressed.predicates import (
        is_dataclass_type,
        is_datetime_type,
        is_enum_type,
        is_fspath_type,
        is_ipaddress_type,
        is_namedtuple_type,
        is_uuid_type,
    )

    return {
        is_ipaddress_type: dump_simple_scalar,
        is_uuid_type: dump_simple_scalar,
        is_enum_type: dump_enum,
        is_datetime_type: dump_datetime,
        is_dataclass_type: dump_dataclass,
        is_namedtuple_type: dump_namedtuple,
        is_fspath_type: dump_fspath,
    }


class Dumper:
    def __init__(
        self,
        *,
        hide_defaults: bool = False,
        default_type_dumpers: Mapping[type, DumperFn] | None = None,
        default_type_mappers: Mapping[TypePredicate, DumperFn] | None = None,
        extra_type_dumpers: Mapping[type, DumperFn] | None = None,
        extra_type_mappers: Mapping[TypePredicate, DumperFn] | None = None,
        enable_specialization: bool = False,
        # If set enables alias lookup on fields, for example for dataclasses.
        alias_field: str | None = "alias",
        # Pass custom alias function, for example mapping field names to camelCase using to_camel.
        alias_fn: AliasFn | None = None,
        # Overriding the resolver factory allows for more advanced behavior like disablign caching,
        # or changing the alias resolution behavior entirely.
        alias_resolver_factory: Callable[[AliasFn | None], AliasResolver] | None = None,
    ) -> None:
        # Map a type to its dumper
        if default_type_dumpers is None:
            type_dumpers = _default_type_dumpers()
        else:
            type_dumpers = dict(default_type_dumpers)
        if extra_type_dumpers:
            type_dumpers |= extra_type_dumpers
        self._type_dumpers: dict[type, DumperFn] = type_dumpers

        # Mapping of type predicate to a dumper
        if default_type_mappers is None:
            type_mappers = _default_type_mappers(enable_specialization)
        else:
            type_mappers = dict(default_type_mappers)
        if extra_type_mappers:
            type_mappers |= extra_type_mappers
        self._type_mappers: Mapping[TypePredicate, DumperFn] = type_mappers

        from tressed.alias import (
            AliasResolver,
            compose_alias_fn,
            identity_alias_fn,
            make_maybe_dataclass_alias_fn,
        )

        if alias_field:
            alias_fn = compose_alias_fn(
                make_maybe_dataclass_alias_fn(alias_field=alias_field),
                alias_fn if alias_fn is not None else identity_alias_fn,
            )

        if alias_resolver_factory is None:
            self._alias_resolver = AliasResolver(alias_fn)
        else:
            self._alias_resolver = alias_resolver_factory(alias_fn)
        self.hide_defaults = hide_defaults

    def _resolve_alias(self, type_: type, type_path: TypePath, name: str) -> Alias:
        return self._alias_resolver.resolve(name, type_, type_path)

    def _dump(self, value: Any, type_path: TypePath) -> Dumped:
        type_: type = type(value)
        if (type_dumper := self._type_dumpers.get(type_)) is None:
            for type_predicate, type_dumper in self._type_mappers.items():
                if type_predicate(type_):
                    break
            else:
                type_dumper = None
            if type_dumper is None:
                raise TressedTypeError(value, type_path)

            # Cache lookup for next time
            self._type_dumpers[type_] = type_dumper

        try:
            return type_dumper(value, type_path, self)
        except TressedValueError:
            raise
        except Exception as e:
            # TODO: Different error for load and dump
            error = TressedValueError(value, type_, type_path)
            error.add_note(f"{type(e)}: {e}")
            raise error from e

    def dump(self, value: Any) -> Dumped:
        return self._dump(value, ())



TYPE_CHECKING = False
if TYPE_CHECKING:
    from tressed.dumper.dumper import Dumper
    from tressed.dumper.types import Dumped, DumperFn, DumperProtocol

__all__ = ["Dumper", "Dumped", "DumperProtocol", "DumperFn"]


if not TYPE_CHECKING:

    def __getattr__(name: str):
        match name:
            case "Dumper":
                from tressed.dumper.dumper import Dumper

                return Dumper

            case "Dumped" | "DumperProtocol" | "DumperFn":
                from tressed.dumper import types

                return getattr(types, name)

            case _:
                raise AttributeError(
                    f"Package '{__package__}' has no attribute '{name}'"
                )



import sys

from tressed.exceptions import TressedValueError
from tressed.predicates import get_args, get_origin, is_union_type
from tressed.type_form import type_form_repr

TYPE_CHECKING = False
if TYPE_CHECKING:
    from collections.abc import Iterator
    from typing import Any

    from tressed.loader.types import LoaderProtocol, TypePath
    from tressed.type_form import TypeForm

__all__ = [
    "load_identity",
    "load_simple_scalar",
    "load_float",
    "load_complex",
    "load_dict",
    "load_simple_collection",
    "load_tuple",
    "load_dataclass",
    "load_newtype",
    "load_typeddict",
    "load_namedtuple",
    "load_literal",
    "load_type_alias",
    "load_optional",
    "load_union",
    "load_datetime",
    "load_discriminated_union",
]

if TYPE_CHECKING:
    from enum import Enum, auto

    class _MissingType(Enum):
        _MISSING = auto()

    _MISSING = _MissingType._MISSING

else:
    _MISSING = object()


def _items(value: Any) -> Iterator[tuple[Any, Any]]:
    type_ = type(value)
    if type_ is dict:
        return value.items()
    elif (argparse := sys.modules.get("argparse")) and type_ is argparse.Namespace:
        return vars(value).items()
    else:
        return value.items()


def load_identity[T](
    value: Any, type_form: TypeForm[T], type_path: TypePath, loader: LoaderProtocol
) -> T:
    """
    Pass the value as is, given that it is of the expected type.
    """
    if type(value) is type_form:
        return value
    raise TressedValueError(
        value,
        type_form,
        type_path,
    )


def load_simple_scalar[T](
    value: Any, type_form: TypeForm[T], type_path: TypePath, loader: LoaderProtocol
) -> T:
    """
    Construct an instance of the given type using the value as the sole argument.
    """
    return type_form(value)  # type: ignore[call-arg]


def load_float[T](
    value: Any, type_form: TypeForm[T], type_path: TypePath, loader: LoaderProtocol
) -> T:
    """
    Load float with int promotion.
    """
    type_ = type(value)
    if type_ is float:
        return value
    if type_ is int:
        return type_form(value)  # type: ignore[call-arg]
    raise TressedValueError(
        value,
        type_form,
        type_path,
    )


def load_complex[T](
    value: Any, type_form: TypeForm[T], type_path: TypePath, loader: LoaderProtocol
) -> T:
    """
    Load complex from a sequence a pair of floats.
    """
    real = loader._load(value[0], float, (*type_path, 0))
    imag = loader._load(value[1], float, (*type_path, 1))
    return type_form(real, imag)  # type: ignore[call-arg]


def load_dict[T](
    value: Any, type_form: TypeForm[T], type_path: TypePath, loader: LoaderProtocol
) -> T:
    args = get_args(type_form)

    assert args is not None
    assert len(args) == 2

    key_type, value_type = args
    return {  # type: ignore[return-value]
        loader._load(
            item_key, key_type, (item_path := (*type_path, item_key))
        ): loader._load(item_value, value_type, item_path)
        for item_key, item_value in _items(value)
    }


def load_simple_collection[T](
    value: Any, type_form: TypeForm[T], type_path: TypePath, loader: LoaderProtocol
) -> T:
    origin = get_origin(type_form)
    num_expected_args = 2 if origin is tuple else 1
    args = get_args(type_form)
    if origin is None or args is None or len(args) != num_expected_args:
        raise TressedValueError(
            value,
            type_form,
            type_path,
            f"{type_form_repr(type_form)} is not a homogeneous generic type",
        )

    item_type = args[0]
    return origin(
        loader._load(item, item_type, (*type_path, pos))
        for pos, item in enumerate(value)
    )


def load_tuple[T](
    value: Any, type_form: TypeForm[T], type_path: TypePath, loader: LoaderProtocol
) -> T:
    origin = get_origin(type_form)
    args = get_args(type_form)
    if origin is None or args is None:
        raise TressedValueError(
            value,
            type_form,
            type_path,
            f"{getattr(type_form, '__name__', type_form)} is not a generic type",
        )

    return tuple(  # type: ignore[return-value]
        loader._load(item, args[pos], (*type_path, pos))
        for pos, item in enumerate(value)
    )


def load_dataclass[T](
    value: Any, type_form: TypeForm[T], type_path: TypePath, loader: LoaderProtocol
) -> T:
    from dataclasses import fields

    loaded = {}

    for field in fields(type_form):  # type: ignore[arg-type]
        field_name = field.name
        alias = loader._resolve_alias(type_form, type_path, field_name)
        if not field.init:
            continue
        if (field_value := value.get(alias, _MISSING)) is not _MISSING:
            loaded[field_name] = field_value
    return type_form(**loaded)


def load_newtype[T: TypeForm](
    value: Any, type_form: T, type_path: TypePath, loader: LoaderProtocol
) -> T:
    supertype = getattr(type_form, "__supertype__")
    return loader._load(value, supertype, type_path)


def _is_extra_items_sentinel(value: Any) -> bool:
    if (typing := sys.modules.get("typing")) and hasattr(typing, "NoExtraItems"):
        if value is getattr(typing, "NoExtraItems"):
            return True
    if typing_extensions := sys.modules.get("typing_extensions"):
        if value is typing_extensions.NoExtraItems:
            return True
    return False


def load_typeddict[T](
    value: Any, type_form: TypeForm[T], type_path: TypePath, loader: LoaderProtocol
) -> T:
    values: dict[str, Any] = {}
    required_keys = getattr(type_form, "__required_keys__")
    optional_keys = getattr(type_form, "__optional_keys__")
    valid_keys = required_keys | optional_keys

    closed = getattr(type_form, "__closed__", None)
    extra_items = getattr(type_form, "__extra_items__", _MISSING)
    if _is_extra_items_sentinel(extra_items):
        extra_items = _MISSING

    extra_keys: set[str] | None = None
    for item_key, item_value in _items(value):
        if extra_items is not _MISSING:
            if item_key not in valid_keys:
                values[item_key] = loader._load(
                    item_value, extra_items, (*type_path, item_key)
                )
                continue

        elif closed:
            if item_key not in valid_keys:
                if extra_keys is None:
                    extra_keys = set()
                extra_keys.add(item_key)
                continue

        values[item_key] = item_value

    if missing_keys := (required_keys - values.keys()):
        raise TressedValueError(
            value,
            type_form,
            type_path,
            f"missing required keys {', '.join(map(repr, missing_keys))}: {value}",
        )

    if extra_keys:
        raise TressedValueError(
            value,
            type_form,
            type_path,
            f"extra keys {', '.join(sorted(map(repr, extra_keys)))}: {value}",
        )

    return values  # type: ignore[return-value]


def load_namedtuple[T](
    value: Any, type_form: TypeForm[T], type_path: TypePath, loader: LoaderProtocol
) -> T:
    from typing import get_type_hints

    type_hints = get_type_hints(type_form)
    values: dict[str, Any] = {
        key: loader._load(field_value, type_hints[key], (*type_path, key))
        for key, field_value in _items(value)
    }
    return type_form(**values)


def load_literal[T](
    value: Any, type_form: TypeForm[T], type_path: TypePath, loader: LoaderProtocol
) -> T:
    args = get_args(type_form)
    assert args is not None
    if value in args:
        return value
    raise TressedValueError(
        value,
        type_form,
        type_path,
        f"got value {repr(value)} but expected one of: {', '.join(map(repr, args))}",
    )


def load_type_alias[T](
    value: Any, type_form: TypeForm[T], type_path: TypePath, loader: LoaderProtocol
) -> T:
    evaluated_type = type_form.evaluate_value()  # type: ignore[attr-defined]
    if (num_params := len(type_form.__type_params__)) > 0:
        args = get_args(type_form)
        if args is None or len(args) < num_params:
            raise TressedValueError(
                value,
                type_form,
                type_path,
                "type form should have only concrete type parameters",
            )

        evaluated_type = evaluated_type[*args]

    return loader._load(value, evaluated_type, type_path)


def load_optional[T](
    value: Any, type_form: TypeForm[T], type_path: TypePath, loader: LoaderProtocol
) -> T:
    if value is None:
        return value  # type: ignore[return-value]

    args = get_args(type_form)
    match args:
        case [T, NoneType] if NoneType is type(None):
            return loader._load(value, T, type_path)
        case [NoneType, T] if NoneType is type(None):
            return loader._load(value, T, type_path)
        case [T]:
            return loader._load(value, T, type_path)
        case _:
            assert False, "unreachable"


def load_union[T](
    value: Any, type_form: TypeForm[T], type_path: TypePath, loader: LoaderProtocol
) -> T:
    args = get_args(type_form)
    assert args, "unreachable"

    errors = None
    for arg in args:
        try:
            return loader._load(value, arg, type_path)
        except TressedValueError as error:
            if errors is None:
                errors = []
            errors.append(error)

    assert args, "unreachable"
    raise TressedValueError(
        value,
        type_form,
        type_path,
        exceptions=errors,
    )


def load_datetime[T](
    value: Any, type_form: TypeForm[T], type_path: TypePath, loader: LoaderProtocol
) -> T:
    return type_form.fromisoformat(value)  # type: ignore[attr-defined]


def load_discriminated_union[T](
    value: Any, type_form: TypeForm[T], type_path: TypePath, loader: LoaderProtocol
) -> T:
    from tressed.discriminated_union import Discriminator

    args = get_args(type_form)
    assert args is not None, "unreachable"

    union_type = next(arg for arg in args if is_union_type(arg))
    union_args = get_args(union_type)
    assert union_args

    discriminator = next(
        metadata
        for metadata in type_form.__metadata__  # type: ignore[attr-defined]
        if isinstance(metadata, Discriminator)
    )

    matched_type_form = discriminator.match(value, *union_args)
    if matched_type_form is None:
        raise TressedValueError(
            value,
            type_form,
            type_path,
            "value did not match discriminated union discriminant",
        )
    return loader._load(value, matched_type_form, type_path)
